!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

      module sort_mod

!----------------------------------------------------
! This module contains routines for sorting vectors.  It uses the SLATEC
! routines IPSORT, SPSORT and DPSORT obtained from GAMS.
! Also provides IPPERM, SPPERM and DPPERM from GAMS.
!
! The routines have been modified to:
!    - not require XERMSG
!    - subroutine end statements have keyword subroutine
! Changes are flagged with WFM.
!
! A generic interface has been been placed on top of the routines.
!
! Further changes on 11/02/2017 to remove Fortran 2015 depreciated features
!    - change from fixed format to free format by
!      - changing C, c and * comment indicators in column 1 with !
!      - change continuation lines from character in column 6 to & at end
!    - replace labelled DO statements with DO ENDDO and GOTO label where
!      label is the end of the do loop with CYCLE
!----------------------------------------------------

      interface sort
         module procedure ipsort
         module procedure spsort
         module procedure dpsort
      end interface

      interface perm
         module procedure ipperm
         module procedure spperm
         module procedure dpperm
      end interface

      contains

!DECK IPSORT
      SUBROUTINE IPSORT (IX, N, IPERM, KFLAG, IER)
!***BEGIN PROLOGUE  IPSORT
!***PURPOSE  Return the permutation vector generated by sorting a given
!            array and, optionally, rearrange the elements of the array.
!            The array may be sorted in increasing or decreasing order.
!            A slightly modified quicksort algorithm is used.
!***LIBRARY   SLATEC
!***CATEGORY  N6A1A, N6A2A
!***TYPE      INTEGER (SPSORT-S, DPSORT-D, IPSORT-I, HPSORT-H)
!***KEYWORDS  NUMBER SORTING, PASSIVE SORTING, SINGLETON QUICKSORT, SORT
!***AUTHOR  Jones, R. E., (SNLA)
!           Kahaner, D. K., (NBS)
!           Rhoads, G. S., (NBS)
!           Wisniewski, J. A., (SNLA)
!***DESCRIPTION
!
!   IPSORT returns the permutation vector IPERM generated by sorting
!   the array IX and, optionally, rearranges the values in IX.  IX may
!   be sorted in increasing or decreasing order.  A slightly modified
!   quicksort algorithm is used.
!
!   IPERM is such that IX(IPERM(I)) is the Ith value in the
!   rearrangement of IX.  IPERM may be applied to another array by
!   calling IPPERM, SPPERM, DPPERM or HPPERM.
!
!   The main difference between IPSORT and its active sorting equivalent
!   ISORT is that the data are referenced indirectly rather than
!   directly.  Therefore, IPSORT should require approximately twice as
!   long to execute as ISORT.  However, IPSORT is more general.
!
!   Description of Parameters
!      IX - input/output -- integer array of values to be sorted.
!           If ABS(KFLAG) = 2, then the values in IX will be
!           rearranged on output; otherwise, they are unchanged.
!      N  - input -- number of values in array IX to be sorted.
!      IPERM - output -- permutation array such that IPERM(I) is the
!              index of the value in the original order of the
!              IX array that is in the Ith location in the sorted
!              order.
!      KFLAG - input -- control parameter:
!            =  2  means return the permutation vector resulting from
!                  sorting IX in increasing order and sort IX also.
!            =  1  means return the permutation vector resulting from
!                  sorting IX in increasing order and do not sort IX.
!            = -1  means return the permutation vector resulting from
!                  sorting IX in decreasing order and do not sort IX.
!            = -2  means return the permutation vector resulting from
!                  sorting IX in decreasing order and sort IX also.
!      IER - output -- error indicator:
!          =  0  if no error,
!          =  1  if N is zero or negative,
!          =  2  if KFLAG is not 2, 1, -1, or -2.
!***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
!                 for sorting with minimal storage, Communications of
!                 the ACM, 12, 3 (1969), pp. 185-187.
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   761101  DATE WRITTEN
!   761118  Modified by John A. Wisniewski to use the Singleton
!           quicksort algorithm.
!   810801  Further modified by David K. Kahaner.
!   870423  Modified by Gregory S. Rhoads for passive sorting with the
!           option for the rearrangement of the original data.
!   890620  Algorithm for rearranging the data vector corrected by R.
!           Boisvert.
!   890622  Prologue upgraded to Version 4.0 style by D. Lozier.
!   891128  Error when KFLAG.LT.0 and N=1 corrected by R. Boisvert.
!   920507  Modified by M. McClain to revise prologue text.
!   920818  Declarations section rebuilt and code restructured to use
!           IF-THEN-ELSE-ENDIF.  (SMR, WRB)
!***END PROLOGUE  IPSORT
!     .. Scalar Arguments ..
      INTEGER IER, KFLAG, N
!     .. Array Arguments ..
      INTEGER IPERM(*), IX(*)
!     .. Local Scalars ..
      REAL R
      INTEGER I, IJ, INDX, INDX0, ISTRT, ITEMP, J, K, KK, L, LM, LMT, M, &
              NN
!     .. Local Arrays ..
      INTEGER IL(21), IU(21)
!     .. External Subroutines ..
! WFM      EXTERNAL XERMSG
!     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
!***FIRST EXECUTABLE STATEMENT  IPSORT
      IER = 0
      NN = N
      IF (NN .LT. 1) THEN
         IER = 1
! WFM         CALL XERMSG ('SLATEC', 'IPSORT',
! WFM     +    'The number of values to be sorted, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
      KK = ABS(KFLAG)
      IF (KK.NE.1 .AND. KK.NE.2) THEN
         IER = 2
! WFM         CALL XERMSG ('SLATEC', 'IPSORT',
! WFM     +    'The sort control parameter, KFLAG, is not 2, 1, -1, or -2.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     Initialize permutation vector
!
      DO I=1,NN
         IPERM(I) = I
      END DO
   10 CONTINUE
!
!     Return if only one value is to be sorted
!
      IF (NN .EQ. 1) RETURN
!
!     Alter array IX to get decreasing order if needed
!
      IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            IX(I) = -IX(I)
         END DO
   20    CONTINUE
      ENDIF
!
!     Sort IX only
!
      M = 1
      I = 1
      J = NN
      R = .375E0
!
   30 IF (I .EQ. J) GO TO 80
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
!
   40 K = I
!
!     Select a central element of the array and save it in location L
!
      IJ = I + INT((J-I)*R)
      LM = IPERM(IJ)
!
!     If first element of array is greater than LM, interchange with LM
!
      IF (IX(IPERM(I)) .GT. IX(LM)) THEN
         IPERM(IJ) = IPERM(I)
         IPERM(I) = LM
         LM = IPERM(IJ)
      ENDIF
      L = J
!
!     If last element of array is less than LM, interchange with LM
!
      IF (IX(IPERM(J)) .LT. IX(LM)) THEN
         IPERM(IJ) = IPERM(J)
         IPERM(J) = LM
         LM = IPERM(IJ)
!
!        If first element of array is greater than LM, interchange
!        with LM
!
         IF (IX(IPERM(I)) .GT. IX(LM)) THEN
            IPERM(IJ) = IPERM(I)
            IPERM(I) = LM
            LM = IPERM(IJ)
         ENDIF
      ENDIF
      GO TO 60
   50 LMT = IPERM(L)
      IPERM(L) = IPERM(K)
      IPERM(K) = LMT
!
!     Find an element in the second half of the array which is smaller
!     than LM
!
   60 L = L-1
      IF (IX(IPERM(L)) .GT. IX(LM)) GO TO 60
!
!     Find an element in the first half of the array which is greater
!     than LM
!
   70 K = K+1
      IF (IX(IPERM(K)) .LT. IX(LM)) GO TO 70
!
!     Interchange these elements
!
      IF (K .LE. L) GO TO 50
!
!     Save upper and lower subscripts of the array yet to be sorted
!
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 90
!
!     Begin again on another portion of the unsorted array
!
   80 M = M-1
      IF (M .EQ. 0) GO TO 120
      I = IL(M)
      J = IU(M)
!
   90 IF (J-I .GE. 1) GO TO 40
      IF (I .EQ. 1) GO TO 30
      I = I-1
!
  100 I = I+1
      IF (I .EQ. J) GO TO 80
      LM = IPERM(I+1)
      IF (IX(IPERM(I)) .LE. IX(LM)) GO TO 100
      K = I
!
  110 IPERM(K+1) = IPERM(K)
      K = K-1
!
      IF (IX(LM) .LT. IX(IPERM(K))) GO TO 110
      IPERM(K+1) = LM
      GO TO 100
!
!     Clean up
!
  120 IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            IX(I) = -IX(I)
         END DO
  130    CONTINUE
      ENDIF
!
!     Rearrange the values of IX if desired
!
      IF (KK .EQ. 2) THEN
!
!        Use the IPERM vector as a flag.
!        If IPERM(I) < 0, then the I-th value is in correct location
!
         DO ISTRT=1,NN
            IF (IPERM(ISTRT) .GE. 0)  THEN
               INDX = ISTRT
               INDX0 = INDX
               ITEMP = IX(ISTRT)
  140          IF (IPERM(INDX) .GT. 0)  THEN
                  IX(INDX) = IX(IPERM(INDX))
                  INDX0 = INDX
                  IPERM(INDX) = -IPERM(INDX)
                  INDX = ABS(IPERM(INDX))
                  GO TO 140
               ENDIF
               IX(INDX0) = ITEMP
            ENDIF
         END DO
  150    CONTINUE
!
!        Revert the signs of the IPERM values
!
         DO I=1,NN
            IPERM(I) = -IPERM(I)
         END DO
  160    CONTINUE
!
      ENDIF
!
      RETURN
      END SUBROUTINE ! WFM

!DECK SPSORT
      SUBROUTINE SPSORT (X, N, IPERM, KFLAG, IER)
!***BEGIN PROLOGUE  SPSORT
!***PURPOSE  Return the permutation vector generated by sorting a given
!            array and, optionally, rearrange the elements of the array.
!            The array may be sorted in increasing or decreasing order.
!            A slightly modified quicksort algorithm is used.
!***LIBRARY   SLATEC
!***CATEGORY  N6A1B, N6A2B
!***TYPE      SINGLE PRECISION (SPSORT-S, DPSORT-D, IPSORT-I, HPSORT-H)
!***KEYWORDS  NUMBER SORTING, PASSIVE SORTING, SINGLETON QUICKSORT, SORT
!***AUTHOR  Jones, R. E., (SNLA)
!           Rhoads, G. S., (NBS)
!           Wisniewski, J. A., (SNLA)
!***DESCRIPTION
!
!   SPSORT returns the permutation vector IPERM generated by sorting
!   the array X and, optionally, rearranges the values in X.  X may
!   be sorted in increasing or decreasing order.  A slightly modified
!   quicksort algorithm is used.
!
!   IPERM is such that X(IPERM(I)) is the Ith value in the rearrangement
!   of X.  IPERM may be applied to another array by calling IPPERM,
!   SPPERM, DPPERM or HPPERM.
!
!   The main difference between SPSORT and its active sorting equivalent
!   SSORT is that the data are referenced indirectly rather than
!   directly.  Therefore, SPSORT should require approximately twice as
!   long to execute as SSORT.  However, SPSORT is more general.
!
!   Description of Parameters
!      X - input/output -- real array of values to be sorted.
!          If ABS(KFLAG) = 2, then the values in X will be
!          rearranged on output; otherwise, they are unchanged.
!      N - input -- number of values in array X to be sorted.
!      IPERM - output -- permutation array such that IPERM(I) is the
!              index of the value in the original order of the
!              X array that is in the Ith location in the sorted
!              order.
!      KFLAG - input -- control parameter:
!            =  2  means return the permutation vector resulting from
!                  sorting X in increasing order and sort X also.
!            =  1  means return the permutation vector resulting from
!                  sorting X in increasing order and do not sort X.
!            = -1  means return the permutation vector resulting from
!                  sorting X in decreasing order and do not sort X.
!            = -2  means return the permutation vector resulting from
!                  sorting X in decreasing order and sort X also.
!      IER - output -- error indicator:
!          =  0  if no error,
!          =  1  if N is zero or negative,
!          =  2  if KFLAG is not 2, 1, -1, or -2.
!***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
!                 for sorting with minimal storage, Communications of
!                 the ACM, 12, 3 (1969), pp. 185-187.
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   761101  DATE WRITTEN
!   761118  Modified by John A. Wisniewski to use the Singleton
!           quicksort algorithm.
!   870423  Modified by Gregory S. Rhoads for passive sorting with the
!           option for the rearrangement of the original data.
!   890620  Algorithm for rearranging the data vector corrected by R.
!           Boisvert.
!   890622  Prologue upgraded to Version 4.0 style by D. Lozier.
!   891128  Error when KFLAG.LT.0 and N=1 corrected by R. Boisvert.
!   920507  Modified by M. McClain to revise prologue text.
!   920818  Declarations section rebuilt and code restructured to use
!           IF-THEN-ELSE-ENDIF.  (SMR, WRB)
!***END PROLOGUE  SPSORT
!     .. Scalar Arguments ..
      INTEGER IER, KFLAG, N
!     .. Array Arguments ..
      REAL X(*)
      INTEGER IPERM(*)
!     .. Local Scalars ..
      REAL R, TEMP
      INTEGER I, IJ, INDX, INDX0, ISTRT, J, K, KK, L, LM, LMT, M, NN
!     .. Local Arrays ..
      INTEGER IL(21), IU(21)
!     .. External Subroutines ..
! WFM      EXTERNAL XERMSG
!     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
!***FIRST EXECUTABLE STATEMENT  SPSORT
      IER = 0
      NN = N
      IF (NN .LT. 1) THEN
         IER = 1
! WFM         CALL XERMSG ('SLATEC', 'SPSORT',
! WFM     +    'The number of values to be sorted, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
      KK = ABS(KFLAG)
      IF (KK.NE.1 .AND. KK.NE.2) THEN
         IER = 2
! WFM         CALL XERMSG ('SLATEC', 'SPSORT',
! WFM     +    'The sort control parameter, KFLAG, is not 2, 1, -1, or -2.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     Initialize permutation vector
!
      DO I=1,NN
         IPERM(I) = I
      END DO
   10 CONTINUE
!
!     Return if only one value is to be sorted
!
      IF (NN .EQ. 1) RETURN
!
!     Alter array X to get decreasing order if needed
!
      IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            X(I) = -X(I)
         END DO
   20    CONTINUE
      ENDIF
!
!     Sort X only
!
      M = 1
      I = 1
      J = NN
      R = .375E0
!
   30 IF (I .EQ. J) GO TO 80
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
!
   40 K = I
!
!     Select a central element of the array and save it in location L
!
      IJ = I + INT((J-I)*R)
      LM = IPERM(IJ)
!
!     If first element of array is greater than LM, interchange with LM
!
      IF (X(IPERM(I)) .GT. X(LM)) THEN
         IPERM(IJ) = IPERM(I)
         IPERM(I) = LM
         LM = IPERM(IJ)
      ENDIF
      L = J
!
!     If last element of array is less than LM, interchange with LM
!
      IF (X(IPERM(J)) .LT. X(LM)) THEN
         IPERM(IJ) = IPERM(J)
         IPERM(J) = LM
         LM = IPERM(IJ)
!
!        If first element of array is greater than LM, interchange
!        with LM
!
         IF (X(IPERM(I)) .GT. X(LM)) THEN
            IPERM(IJ) = IPERM(I)
            IPERM(I) = LM
            LM = IPERM(IJ)
         ENDIF
      ENDIF
      GO TO 60
   50 LMT = IPERM(L)
      IPERM(L) = IPERM(K)
      IPERM(K) = LMT
!
!     Find an element in the second half of the array which is smaller
!     than LM
!
   60 L = L-1
      IF (X(IPERM(L)) .GT. X(LM)) GO TO 60
!
!     Find an element in the first half of the array which is greater
!     than LM
!
   70 K = K+1
      IF (X(IPERM(K)) .LT. X(LM)) GO TO 70
!
!     Interchange these elements
!
      IF (K .LE. L) GO TO 50
!
!     Save upper and lower subscripts of the array yet to be sorted
!
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 90
!
!     Begin again on another portion of the unsorted array
!
   80 M = M-1
      IF (M .EQ. 0) GO TO 120
      I = IL(M)
      J = IU(M)
!
   90 IF (J-I .GE. 1) GO TO 40
      IF (I .EQ. 1) GO TO 30
      I = I-1
!
  100 I = I+1
      IF (I .EQ. J) GO TO 80
      LM = IPERM(I+1)
      IF (X(IPERM(I)) .LE. X(LM)) GO TO 100
      K = I
!
  110 IPERM(K+1) = IPERM(K)
      K = K-1
!
      IF (X(LM) .LT. X(IPERM(K))) GO TO 110
      IPERM(K+1) = LM
      GO TO 100
!
!     Clean up
!
  120 IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            X(I) = -X(I)
         END DO
  130    CONTINUE
      ENDIF
!
!     Rearrange the values of X if desired
!
      IF (KK .EQ. 2) THEN
!
!        Use the IPERM vector as a flag.
!        If IPERM(I) < 0, then the I-th value is in correct location
!
         DO ISTRT=1,NN
            IF (IPERM(ISTRT) .GE. 0) THEN
               INDX = ISTRT
               INDX0 = INDX
               TEMP = X(ISTRT)
  140          IF (IPERM(INDX) .GT. 0) THEN
                  X(INDX) = X(IPERM(INDX))
                  INDX0 = INDX
                  IPERM(INDX) = -IPERM(INDX)
                  INDX = ABS(IPERM(INDX))
                  GO TO 140
               ENDIF
               X(INDX0) = TEMP
            ENDIF
         END DO
  150    CONTINUE
!
!        Revert the signs of the IPERM values
!
         DO I=1,NN
            IPERM(I) = -IPERM(I)
         END DO
  160    CONTINUE
!
      ENDIF
!
      RETURN
      END SUBROUTINE ! WFM

!DECK DPSORT
      SUBROUTINE DPSORT (DX, N, IPERM, KFLAG, IER)
!***BEGIN PROLOGUE  DPSORT
!***PURPOSE  Return the permutation vector generated by sorting a given
!            array and, optionally, rearrange the elements of the array.
!            The array may be sorted in increasing or decreasing order.
!            A slightly modified quicksort algorithm is used.
!***LIBRARY   SLATEC
!***CATEGORY  N6A1B, N6A2B
!***TYPE      DOUBLE PRECISION (SPSORT-S, DPSORT-D, IPSORT-I, HPSORT-H)
!***KEYWORDS  NUMBER SORTING, PASSIVE SORTING, SINGLETON QUICKSORT, SORT
!***AUTHOR  Jones, R. E., (SNLA)
!           Rhoads, G. S., (NBS)
!           Wisniewski, J. A., (SNLA)
!***DESCRIPTION
!
!   DPSORT returns the permutation vector IPERM generated by sorting
!   the array DX and, optionally, rearranges the values in DX.  DX may
!   be sorted in increasing or decreasing order.  A slightly modified
!   quicksort algorithm is used.
!
!   IPERM is such that DX(IPERM(I)) is the Ith value in the
!   rearrangement of DX.  IPERM may be applied to another array by
!   calling IPPERM, SPPERM, DPPERM or HPPERM.
!
!   The main difference between DPSORT and its active sorting equivalent
!   DSORT is that the data are referenced indirectly rather than
!   directly.  Therefore, DPSORT should require approximately twice as
!   long to execute as DSORT.  However, DPSORT is more general.
!
!   Description of Parameters
!      DX - input/output -- double precision array of values to be
!           sorted.  If ABS(KFLAG) = 2, then the values in DX will be
!           rearranged on output; otherwise, they are unchanged.
!      N  - input -- number of values in array DX to be sorted.
!      IPERM - output -- permutation array such that IPERM(I) is the
!              index of the value in the original order of the
!              DX array that is in the Ith location in the sorted
!              order.
!      KFLAG - input -- control parameter:
!            =  2  means return the permutation vector resulting from
!                  sorting DX in increasing order and sort DX also.
!            =  1  means return the permutation vector resulting from
!                  sorting DX in increasing order and do not sort DX.
!            = -1  means return the permutation vector resulting from
!                  sorting DX in decreasing order and do not sort DX.
!            = -2  means return the permutation vector resulting from
!                  sorting DX in decreasing order and sort DX also.
!      IER - output -- error indicator:
!          =  0  if no error,
!          =  1  if N is zero or negative,
!          =  2  if KFLAG is not 2, 1, -1, or -2.
!***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
!                 for sorting with minimal storage, Communications of
!                 the ACM, 12, 3 (1969), pp. 185-187.
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   761101  DATE WRITTEN
!   761118  Modified by John A. Wisniewski to use the Singleton
!           quicksort algorithm.
!   870423  Modified by Gregory S. Rhoads for passive sorting with the
!           option for the rearrangement of the original data.
!   890619  Double precision version of SPSORT created by D. W. Lozier.
!   890620  Algorithm for rearranging the data vector corrected by R.
!           Boisvert.
!   890622  Prologue upgraded to Version 4.0 style by D. Lozier.
!   891128  Error when KFLAG.LT.0 and N=1 corrected by R. Boisvert.
!   920507  Modified by M. McClain to revise prologue text.
!   920818  Declarations section rebuilt and code restructured to use
!           IF-THEN-ELSE-ENDIF.  (SMR, WRB)
!***END PROLOGUE  DPSORT
!     .. Scalar Arguments ..
      INTEGER IER, KFLAG, N
!     .. Array Arguments ..
      DOUBLE PRECISION DX(*)
      INTEGER IPERM(*)
!     .. Local Scalars ..
      DOUBLE PRECISION R, TEMP
      INTEGER I, IJ, INDX, INDX0, ISTRT, J, K, KK, L, LM, LMT, M, NN
!     .. Local Arrays ..
      INTEGER IL(21), IU(21)
!     .. External Subroutines ..
! WFM      EXTERNAL XERMSG
!     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
!***FIRST EXECUTABLE STATEMENT  DPSORT
      IER = 0
      NN = N
      IF (NN .LT. 1) THEN
         IER = 1
! WFM         CALL XERMSG ('SLATEC', 'DPSORT',
! WFM     +    'The number of values to be sorted, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
      KK = ABS(KFLAG)
      IF (KK.NE.1 .AND. KK.NE.2) THEN
         IER = 2
! WFM         CALL XERMSG ('SLATEC', 'DPSORT',
! WFM     +    'The sort control parameter, KFLAG, is not 2, 1, -1, or -2.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     Initialize permutation vector
!
      DO I=1,NN
         IPERM(I) = I
      END DO
   10 CONTINUE
!
!     Return if only one value is to be sorted
!
      IF (NN .EQ. 1) RETURN
!
!     Alter array DX to get decreasing order if needed
!
      IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            DX(I) = -DX(I)
         END DO
   20    CONTINUE
      ENDIF
!
!     Sort DX only
!
      M = 1
      I = 1
      J = NN
      R = .375D0
!
   30 IF (I .EQ. J) GO TO 80
      IF (R .LE. 0.5898437D0) THEN
         R = R+3.90625D-2
      ELSE
         R = R-0.21875D0
      ENDIF
!
   40 K = I
!
!     Select a central element of the array and save it in location L
!
      IJ = I + INT((J-I)*R)
      LM = IPERM(IJ)
!
!     If first element of array is greater than LM, interchange with LM
!
      IF (DX(IPERM(I)) .GT. DX(LM)) THEN
         IPERM(IJ) = IPERM(I)
         IPERM(I) = LM
         LM = IPERM(IJ)
      ENDIF
      L = J
!
!     If last element of array is less than LM, interchange with LM
!
      IF (DX(IPERM(J)) .LT. DX(LM)) THEN
         IPERM(IJ) = IPERM(J)
         IPERM(J) = LM
         LM = IPERM(IJ)
!
!        If first element of array is greater than LM, interchange
!        with LM
!
         IF (DX(IPERM(I)) .GT. DX(LM)) THEN
            IPERM(IJ) = IPERM(I)
            IPERM(I) = LM
            LM = IPERM(IJ)
         ENDIF
      ENDIF
      GO TO 60
   50 LMT = IPERM(L)
      IPERM(L) = IPERM(K)
      IPERM(K) = LMT
!
!     Find an element in the second half of the array which is smaller
!     than LM
!
   60 L = L-1
      IF (DX(IPERM(L)) .GT. DX(LM)) GO TO 60
!
!     Find an element in the first half of the array which is greater
!     than LM
!
   70 K = K+1
      IF (DX(IPERM(K)) .LT. DX(LM)) GO TO 70
!
!     Interchange these elements
!
      IF (K .LE. L) GO TO 50
!
!     Save upper and lower subscripts of the array yet to be sorted
!
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 90
!
!     Begin again on another portion of the unsorted array
!
   80 M = M-1
      IF (M .EQ. 0) GO TO 120
      I = IL(M)
      J = IU(M)
!
   90 IF (J-I .GE. 1) GO TO 40
      IF (I .EQ. 1) GO TO 30
      I = I-1
!
  100 I = I+1
      IF (I .EQ. J) GO TO 80
      LM = IPERM(I+1)
      IF (DX(IPERM(I)) .LE. DX(LM)) GO TO 100
      K = I
!
  110 IPERM(K+1) = IPERM(K)
      K = K-1
      IF (DX(LM) .LT. DX(IPERM(K))) GO TO 110
      IPERM(K+1) = LM
      GO TO 100
!
!     Clean up
!
  120 IF (KFLAG .LE. -1) THEN
         DO I=1,NN
            DX(I) = -DX(I)
         END DO
  130    CONTINUE
      ENDIF
!
!     Rearrange the values of DX if desired
!
      IF (KK .EQ. 2) THEN
!
!        Use the IPERM vector as a flag.
!        If IPERM(I) < 0, then the I-th value is in correct location
!
         DO ISTRT=1,NN
            IF (IPERM(ISTRT) .GE. 0) THEN
               INDX = ISTRT
               INDX0 = INDX
               TEMP = DX(ISTRT)
  140          IF (IPERM(INDX) .GT. 0) THEN
                  DX(INDX) = DX(IPERM(INDX))
                  INDX0 = INDX
                  IPERM(INDX) = -IPERM(INDX)
                  INDX = ABS(IPERM(INDX))
                  GO TO 140
               ENDIF
               DX(INDX0) = TEMP
            ENDIF
         END DO
  150    CONTINUE
!
!        Revert the signs of the IPERM values
!
         DO I=1,NN
            IPERM(I) = -IPERM(I)
         END DO
  160    CONTINUE
!
      ENDIF
!
      RETURN
      END SUBROUTINE ! WFM

!DECK IPPERM
      SUBROUTINE IPPERM (IX, N, IPERM, IER)
!***BEGIN PROLOGUE  IPPERM
!***PURPOSE  Rearrange a given array according to a prescribed
!            permutation vector.
!***LIBRARY   SLATEC
!***CATEGORY  N8
!***TYPE      INTEGER (SPPERM-S, DPPERM-D, IPPERM-I, HPPERM-H)
!***KEYWORDS  APPLICATION OF PERMUTATION TO DATA VECTOR
!***AUTHOR  McClain, M. A., (NIST)
!           Rhoads, G. S., (NBS)
!***DESCRIPTION
!
!         IPPERM rearranges the data vector IX according to the
!         permutation IPERM: IX(I) <--- IX(IPERM(I)).  IPERM could come
!         from one of the sorting routines IPSORT, SPSORT, DPSORT or
!         HPSORT.
!
!     Description of Parameters
!         IX - input/output -- integer array of values to be rearranged.
!         N - input -- number of values in integer array IX.
!         IPERM - input -- permutation vector.
!         IER - output -- error indicator:
!             =  0  if no error,
!             =  1  if N is zero or negative,
!             =  2  if IPERM is not a valid permutation.
!
!***REFERENCES  (NONE)
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   900618  DATE WRITTEN
!   920507  Modified by M. McClain to revise prologue text.
!***END PROLOGUE  IPPERM
      INTEGER IX(*), N, IPERM(*), I, IER, INDX, INDX0, ITEMP, ISTRT
!***FIRST EXECUTABLE STATEMENT  IPPERM
      IER=0
      IF(N.LT.1)THEN
         IER=1
! WFM         CALL XERMSG ('SLATEC', 'IPPERM',
! WFM     +    'The number of values to be rearranged, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     CHECK WHETHER IPERM IS A VALID PERMUTATION
!
      DO I=1,N
         INDX=ABS(IPERM(I))
         IF((INDX.GE.1).AND.(INDX.LE.N))THEN
            IF(IPERM(INDX).GT.0)THEN
               IPERM(INDX)=-IPERM(INDX)
               CYCLE
            ENDIF
         ENDIF
         IER=2
! WFM         CALL XERMSG ('SLATEC', 'IPPERM',
! WFM     +    'The permutation vector, IPERM, is not valid.', IER, 1)
         RETURN
      END DO
  100 CONTINUE
!
!     REARRANGE THE VALUES OF IX
!
!     USE THE IPERM VECTOR AS A FLAG.
!     IF IPERM(I) > 0, THEN THE I-TH VALUE IS IN CORRECT LOCATION
!
      DO ISTRT = 1 , N
         IF (IPERM(ISTRT) .GT. 0) CYCLE
         INDX = ISTRT
         INDX0 = INDX
         ITEMP = IX(ISTRT)
  320    CONTINUE
         IF (IPERM(INDX) .GE. 0) GOTO 325
            IX(INDX) = IX(-IPERM(INDX))
            INDX0 = INDX
            IPERM(INDX) = -IPERM(INDX)
            INDX = IPERM(INDX)
            GOTO 320
  325    CONTINUE
         IX(INDX0) = ITEMP
      END DO
  330 CONTINUE
!
      RETURN
      END SUBROUTINE ! WFM

!DECK DPPERM
      SUBROUTINE DPPERM (DX, N, IPERM, IER)
!***BEGIN PROLOGUE  DPPERM
!***PURPOSE  Rearrange a given array according to a prescribed
!            permutation vector.
!***LIBRARY   SLATEC
!***CATEGORY  N8
!***TYPE      DOUBLE PRECISION (SPPERM-S, DPPERM-D, IPPERM-I, HPPERM-H)
!***KEYWORDS  PERMUTATION, REARRANGEMENT
!***AUTHOR  McClain, M. A., (NIST)
!           Rhoads, G. S., (NBS)
!***DESCRIPTION
!
!         DPPERM rearranges the data vector DX according to the
!         permutation IPERM: DX(I) <--- DX(IPERM(I)).  IPERM could come
!         from one of the sorting routines IPSORT, SPSORT, DPSORT or
!         HPSORT.
!
!     Description of Parameters
!         DX - input/output -- double precision array of values to be
!                   rearranged.
!         N - input -- number of values in double precision array DX.
!         IPERM - input -- permutation vector.
!         IER - output -- error indicator:
!             =  0  if no error,
!             =  1  if N is zero or negative,
!             =  2  if IPERM is not a valid permutation.
!
!***REFERENCES  (NONE)
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   901004  DATE WRITTEN
!   920507  Modified by M. McClain to revise prologue text.
!***END PROLOGUE  DPPERM
      INTEGER N, IPERM(*), I, IER, INDX, INDX0, ISTRT
      DOUBLE PRECISION DX(*), DTEMP
!***FIRST EXECUTABLE STATEMENT  DPPERM
      IER=0
      IF(N.LT.1)THEN
         IER=1
! WFM         CALL XERMSG ('SLATEC', 'DPPERM',
! WFM     +    'The number of values to be rearranged, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     CHECK WHETHER IPERM IS A VALID PERMUTATION
!
      DO I=1,N
         INDX=ABS(IPERM(I))
         IF((INDX.GE.1).AND.(INDX.LE.N))THEN
            IF(IPERM(INDX).GT.0)THEN
               IPERM(INDX)=-IPERM(INDX)
               CYCLE
            ENDIF
         ENDIF
         IER=2
! WFM         CALL XERMSG ('SLATEC', 'DPPERM',
! WFM     +    'The permutation vector, IPERM, is not valid.', IER, 1)
         RETURN
      END DO
  100 CONTINUE
!
!     REARRANGE THE VALUES OF DX
!
!     USE THE IPERM VECTOR AS A FLAG.
!     IF IPERM(I) > 0, THEN THE I-TH VALUE IS IN CORRECT LOCATION
!
      DO ISTRT = 1 , N
         IF (IPERM(ISTRT) .GT. 0) CYCLE
         INDX = ISTRT
         INDX0 = INDX
         DTEMP = DX(ISTRT)
  320    CONTINUE
         IF (IPERM(INDX) .GE. 0) GOTO 325
            DX(INDX) = DX(-IPERM(INDX))
            INDX0 = INDX
            IPERM(INDX) = -IPERM(INDX)
            INDX = IPERM(INDX)
            GOTO 320
  325    CONTINUE
         DX(INDX0) = DTEMP
      END DO
  330 CONTINUE
!
      RETURN
      END SUBROUTINE ! WFM

!DECK SPPERM
      SUBROUTINE SPPERM (X, N, IPERM, IER)
!***BEGIN PROLOGUE  SPPERM
!***PURPOSE  Rearrange a given array according to a prescribed
!            permutation vector.
!***LIBRARY   SLATEC
!***CATEGORY  N8
!***TYPE      SINGLE PRECISION (SPPERM-S, DPPERM-D, IPPERM-I, HPPERM-H)
!***KEYWORDS  APPLICATION OF PERMUTATION TO DATA VECTOR
!***AUTHOR  McClain, M. A., (NIST)
!           Rhoads, G. S., (NBS)
!***DESCRIPTION
!
!         SPPERM rearranges the data vector X according to the
!         permutation IPERM: X(I) <--- X(IPERM(I)).  IPERM could come
!         from one of the sorting routines IPSORT, SPSORT, DPSORT or
!         HPSORT.
!
!     Description of Parameters
!         X - input/output -- real array of values to be rearranged.
!         N - input -- number of values in real array X.
!         IPERM - input -- permutation vector.
!         IER - output -- error indicator:
!             =  0  if no error,
!             =  1  if N is zero or negative,
!             =  2  if IPERM is not a valid permutation.
!
!***REFERENCES  (NONE)
!***ROUTINES CALLED  XERMSG
!***REVISION HISTORY  (YYMMDD)
!   901004  DATE WRITTEN
!   920507  Modified by M. McClain to revise prologue text.
!***END PROLOGUE  SPPERM
      INTEGER N, IPERM(*), I, IER, INDX, INDX0, ISTRT
      REAL X(*), TEMP
!***FIRST EXECUTABLE STATEMENT  SPPERM
      IER=0
      IF(N.LT.1)THEN
         IER=1
! WFM         CALL XERMSG ('SLATEC', 'SPPERM',
! WFM     +    'The number of values to be rearranged, N, is not positive.',
! WFM     +    IER, 1)
         RETURN
      ENDIF
!
!     CHECK WHETHER IPERM IS A VALID PERMUTATION
!
      DO I=1,N
         INDX=ABS(IPERM(I))
         IF((INDX.GE.1).AND.(INDX.LE.N))THEN
            IF(IPERM(INDX).GT.0)THEN
               IPERM(INDX)=-IPERM(INDX)
               CYCLE
            ENDIF
         ENDIF
         IER=2
! WFM         CALL XERMSG ('SLATEC', 'SPPERM',
! WFM     +    'The permutation vector, IPERM, is not valid.', IER, 1)
         RETURN
      END DO
  100 CONTINUE
!
!     REARRANGE THE VALUES OF X
!
!     USE THE IPERM VECTOR AS A FLAG.
!     IF IPERM(I) > 0, THEN THE I-TH VALUE IS IN CORRECT LOCATION
!
      DO ISTRT = 1 , N
         IF (IPERM(ISTRT) .GT. 0) CYCLE
         INDX = ISTRT
         INDX0 = INDX
         TEMP = X(ISTRT)
  320    CONTINUE
         IF (IPERM(INDX) .GE. 0) GOTO 325
            X(INDX) = X(-IPERM(INDX))
            INDX0 = INDX
            IPERM(INDX) = -IPERM(INDX)
            INDX = IPERM(INDX)
            GOTO 320
  325    CONTINUE
         X(INDX0) = TEMP
      END DO
  330 CONTINUE
!
      RETURN
      END SUBROUTINE ! WFM

      end module sort_mod
